from __future__ import annotations

import logging
import os
from typing import Optional, Union, cast

from kraken.core import Project, Property, Supplier, Task, TaskStatus
from nr.python.environment.virtualenv import get_current_venv

from ..buildsystem import PythonBuildSystem
from ..settings import python_settings

logger = logging.getLogger(__name__)


class InstallTask(Task):
    build_system: Property[Optional[PythonBuildSystem]]
    always_use_managed_env: Property[bool]

    def prepare(self) -> TaskStatus | None:
        venv = get_current_venv(os.environ)
        if not self.always_use_managed_env.get() and venv:
            return TaskStatus.skipped("using current virtual env (%s)" % venv.path)
        build_system = self.build_system.get()
        if not build_system:
            return TaskStatus.skipped("no Python build system configured")
        if not build_system.supports_managed_environments():
            return TaskStatus.skipped(
                "current build system does not supported managed environment (%s)" % type(build_system).__name__
            )
        managed_environment = build_system.get_managed_environment()
        if managed_environment.exists():
            return TaskStatus.skipped("managed environment exists (%s)" % managed_environment.get_path())
        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        build_system = self.build_system.get()
        if not build_system:
            logger.error("no build system configured")
            return TaskStatus.failed("no build system configured")
        managed_environment = build_system.get_managed_environment()
        managed_environment.install(python_settings(self.project))
        return TaskStatus.succeeded()


def install(project: Project | None = None) -> InstallTask:
    """Get or create the `pythonInstall` task for the given project.

    The install task relies on the build system configured in the Python project settings."""

    name = "pythonInstall"

    project = project or Project.current()
    task = cast(Union[InstallTask, None], project.tasks().get(name))
    if task is None:
        task = project.do(name, InstallTask, default=False)
        task.build_system.set(Supplier.of_callable(lambda: python_settings(project).build_system))
        task.always_use_managed_env.set(Supplier.of_callable(lambda: python_settings(project).always_use_managed_env))

    return task



""" Implements Slap as a Python build system for kraken-std.

Requires at least Slap 1.6.25. """

from __future__ import annotations

import logging
import shutil
import subprocess as sp
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import quote

from kraken.core.utils import NotSet

from . import ManagedEnvironment, PythonBuildSystem

if TYPE_CHECKING:
    from ..settings import PythonSettings

logger = logging.getLogger(__name__)


class SlapPythonBuildSystem(PythonBuildSystem):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory

    def supports_managed_environments(self) -> bool:
        return True

    def get_managed_environment(self) -> ManagedEnvironment:
        return SlapManagedEnvironment(self.project_directory)

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        if as_version is not None:
            # TODO (@NiklasRosenstein): We should find a way to revert the changes to the worktree
            #       that this command does.
            command = ["slap", "release", as_version]
            logger.info("%s", command)
            sp.check_call(command, cwd=self.project_directory)

        with tempfile.TemporaryDirectory() as tempdir:
            command = ["slap", "publish", "--dry", "-b", tempdir]
            sp.check_call(command, cwd=self.project_directory)
            src_files = list(Path(tempdir).iterdir())
            dst_files = [output_directory / path.name for path in src_files]
            for src, dst in zip(src_files, dst_files):
                shutil.move(str(src), dst)
        return dst_files


class SlapManagedEnvironment(ManagedEnvironment):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory
        self._env_path: Path | None | NotSet = NotSet.Value

    def exists(self) -> bool:
        try:
            self.get_path()
            return True
        except RuntimeError:
            return False

    def get_path(self) -> Path:
        if self._env_path is NotSet.Value:
            command = ["slap", "venv", "-p"]
            try:
                self._env_path = Path(
                    sp.check_output(command, cwd=self.project_directory, stderr=sp.DEVNULL).decode().strip()
                )
            except sp.CalledProcessError as exc:
                if exc.returncode != 1:
                    raise
                self._env_path = None
        if self._env_path is None:
            raise RuntimeError("managed environment does not exist")
        return self._env_path

    def install(self, settings: PythonSettings) -> None:
        # Ensure that an environment exists.
        command = ["slap", "venv", "-ac"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)

        # Install into the environment.
        command = ["slap", "install", "--ignore-active-venv", "--link"]
        safe_command = list(command)
        for index in settings.package_indexes.values():
            if index.is_package_source:
                spec = f"name={quote(index.alias)},url={quote(index.index_url)}"
                if index.credentials:
                    spec += f",username={quote(index.credentials[0])},password={quote(index.credentials[1])}"
                safe_spec = spec.replace(quote(index.credentials[1]), "[MASKED]") if index.credentials else spec
                option = "--index" if index.default else "--extra-index"
                command += [option, spec]
                safe_command += [option, safe_spec]

        logger.info("%s", safe_command)
        sp.check_call(command, cwd=self.project_directory)



from __future__ import annotations

import os
import re
import subprocess as sp

from kraken.core.task import TaskStatus
from kraken.core.utils import flatten, not_none

from . import DockerBuildTask


class BuildxBuildTask(DockerBuildTask):
    """Implements building a Docker image with Buildx."""

    def finalize(self) -> None:
        if not self.load.get() and not self.push.get():
            self.logger.info("activating --load because one of --load or --push is necessary with Buildx")
            self.load.set(True)
        return super().finalize()

    def execute(self) -> TaskStatus:
        inspect_response = sp.check_output(["docker", "buildx", "inspect"]).decode()
        if re.search(r"Driver:\s*docker\n", inspect_response) and self.cache_repo.get():
            self.logger.info(
                "creating new Buildx driver, reason: current driver is Docker which does not support cache exports"
            )
            sp.check_call(["docker", "buildx", "create", "--use"])

        command = ["docker", "buildx", "build", str(self.build_context.get().absolute())]
        if self.dockerfile.is_filled():
            command += ["-f", str(self.dockerfile.get().absolute())]
        if self.platform.is_filled():
            command += ["--platform", str(self.platform.get())]
        command += flatten(["--build-arg", f"{k}={v}"] for k, v in self.build_args.get().items())
        command += flatten(["--secret", f"id={k}"] for k in self.secrets.get())
        if self.cache_repo.get():
            # NOTE (@NiklasRosenstein): Buildx does not allow leading underscores, while Kaniko and Artifactory do.
            command += ["--cache-to", f"type=registry,ref={not_none(self.cache_repo.get())}"]
        if not self.cache.get():
            command += ["--no-cache"]
        command += flatten(["--tag", t] for t in self.tags.get())
        if self.push.get():
            command += ["--push"]
        if self.squash.get():
            command += ["--squash"]
        if self.target.get():
            command += ["--target", not_none(self.target.get())]
        if self.image_output_file.get():
            command += ["--output", f"type=tar,dest={self.image_output_file.get()}"]
        if self.load.get():
            command += ["--load"]

        # Buildx will take the secret from the environment variables.
        env = os.environ.copy()
        env.update(self.secrets.get())

        # TODO (@nrosenstein): docker login for auth

        self.logger.info("%s", command)
        result = sp.call(command, env=env, cwd=self.project.directory)
        return TaskStatus.from_exit_code(command, result)



from __future__ import annotations

import os
import subprocess as sp
import tempfile
from pathlib import Path

from kraken.core.project import Project
from kraken.core.property import Property
from kraken.core.task import TaskStatus
from kraken.core.utils import flatten, not_none

from . import DockerBuildTask


class NativeBuildTask(DockerBuildTask):
    """Implements building a Docker image using the native `docker build` command."""

    #: Whether to use Docker Buildkit. Enabled by default.
    native_use_buildkit: Property[bool]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.native_use_buildkit.set(True)

    def finalize(self) -> None:
        if self.cache_repo.get() is not None:
            self.logger.warning("cache_repo is not supported for NativeBuildTask")
        if self.squash.get() is not None:
            self.logger.warning("squash is not supported for NativeBuildTask")
        if not self.load.get():
            self.logger.warning("load is not supported for NativeBuildTask, resulting image will always be loaded")
        if self.push.get() and not self.tags.get():
            raise ValueError(f"{self}.tags cannot be empty if .push is enabled")
        return super().finalize()

    def execute(self) -> TaskStatus:
        command = ["docker", "build", str(self.build_context.get().absolute())]
        if self.dockerfile.is_filled():
            command += ["-f", str(self.dockerfile.get().absolute())]
        if self.platform.is_filled():
            command += ["--platform", str(self.platform.get())]
        command += flatten(["--build-arg", f"{k}={v}"] for k, v in self.build_args.get().items())
        if self.cache_repo.get():
            # NOTE (@NiklasRosenstein): Buildx does not allow leading underscores, while Kaniko and Artifactory do.
            command += ["--cache-from", f"type=registry,ref={not_none(self.cache_repo.get())}"]
        if not self.cache.get():
            command += ["--no-cache"]
        command += flatten(["--tag", t] for t in self.tags.get())
        if self.target.get():
            command += ["--target", not_none(self.target.get())]
        if self.image_output_file.get():
            command += ["--output", f"type=tar,dest={self.image_output_file.get()}"]

        # Buildx will take the secret from the environment variasbles.
        env = os.environ.copy()
        env["DOCKER_BUILDKIT"] = "1" if self.native_use_buildkit.get() else "0"

        # TODO (@nrosenstein): docker login for auth

        with tempfile.TemporaryDirectory() as tempdir:
            for key, value in self.secrets.get().items():
                secret_file = Path(tempdir) / key
                secret_file.write_text(value)
                command += ["--secret", f"id={key},src={secret_file}"]

            self.logger.info("%s", command)
            result = sp.call(command, env=env, cwd=self.project.directory)
            if result != 0:
                return TaskStatus.from_exit_code(command, result)

        if self.push.get():
            command = ["docker", "push"] + self.tags.get()
            self.logger.info("%s", command)
            result = sp.call(command, env=env, cwd=self.project.directory)

        return TaskStatus.succeeded()



from __future__ import annotations

import subprocess as sp
from typing import Optional

from kraken.core import Property, Task, TaskStatus


class CargoClippyTask(Task):
    """Runs `cargo clippy` for linting or applying suggestions."""

    fix: Property[bool] = Property.default(False)
    allow: Property[Optional[str]] = Property.default("staged")

    def execute(self) -> TaskStatus | None:
        command = ["cargo", "clippy"]
        if self.fix.get():
            command += ["--fix"]
            allow = self.allow.get()
            if allow == "staged":
                command += ["--allow-staged"]
            elif allow == "dirty":
                command += ["--allow-dirty"]
            elif allow is not None:
                raise ValueError(f"invalid allow: {allow!r}")
        return TaskStatus.from_exit_code(command, sp.call(command, cwd=self.project.directory))



from __future__ import annotations

import logging
import os
import subprocess as sp
from pathlib import Path

from kraken.core.utils import flatten

logger = logging.getLogger(__name__)


def docker_run(
    image: str,
    args: list[str] | None = None,
    env: dict[str, str] | None = None,
    platform: str | None = None,
    entrypoint: str | None = None,
    interactive: bool = False,
    remove: bool = True,
    volumes: list[str] | None = None,
    workdir: str | None = None,
    cwd: str | None = None,
    environ: dict[str, str] | None = None,
) -> int:
    command = ["docker", "run"]
    command += flatten(("--env", f"{key}={value}") for key, value in (env or {}).items())
    if platform is not None:
        command += ["--platform", platform]
    if entrypoint is not None:
        command += ["--entrypoint", entrypoint]
    if interactive:
        command += ["-it"]
    if remove:
        command += ["--rm"]
    command += flatten(("--volume", volume) for volume in (volumes or []))
    if workdir:
        command += ["--workdir", workdir]
    command += [image]
    command += args or []
    logger.info("%s", command)
    return sp.call(command, cwd=cwd, env={**os.environ, **(environ or {})})


def docker_load(image_file: Path) -> int:
    """Load a docker image archive by file."""

    command = ["docker", "load", "-i", str(image_file)]
    logger.info("%s", command)
    return sp.call(command)



""" Build and publish Helm charts with Kraken. """

from __future__ import annotations

import dataclasses
import urllib.parse
from pathlib import Path

import httpx
from kraken.core import Project, Property, Supplier, Task, TaskStatus

from . import helmapi


@dataclasses.dataclass
class HelmSettings:
    """Project-specific settings for Helm."""

    auth: dict[str, tuple[str, str]] = dataclasses.field(default_factory=dict)
    insecure_registries: set[str] = dataclasses.field(default_factory=set)

    def add_auth(self, host: str, username: str, password: str, insecure: bool = False) -> HelmSettings:
        self.auth[host] = (username, password)
        if insecure:
            self.insecure_registries.add(host)
        return self


def helm_settings(project: Project | None = None) -> HelmSettings:
    """Create or get Helm settings for the project."""

    if project is None:
        from kraken.api import project as _project

        project = _project

    settings = project.find_metadata(HelmSettings)
    if settings is None:
        settings = HelmSettings()
        project.metadata.append(settings)

    return settings


class HelmPackageTask(Task):
    """Packages a Helm chart."""

    # The path to the directory that contains the Helm chart. A relative path is treated relative to
    # the project directory. This property must be set.
    chart_directory: Property[Path]

    # This property specifies the path to the output Helm chart tarball. It can be specified, when the
    # task is created if an explicit output location is desired, otherwise the property will be set
    # when the task was executed and the default output location is in the build directory.
    chart_tarball: Property[Path] = Property.output()

    def execute(self) -> TaskStatus:
        chart_directory = self.project.directory / self.chart_directory.get()
        if self.chart_tarball.is_filled():
            status, output_file = helmapi.helm_package(chart_directory, output_file=self.chart_tarball.get())
        else:
            output_directory = self.project.build_directory / "helm" / self.name
            status, output_file = helmapi.helm_package(chart_directory, output_directory=output_directory)
            if output_file:
                self.chart_tarball.set(output_file)
        if status != 0 or not output_file:
            return TaskStatus.failed()
        return TaskStatus.succeeded()


class HelmPushTask(Task):
    """Pushes a Helm chart to a Helm registry. Supports OCI and HTTP(S) registries."""

    #: The path to the Helm chart package file. This is usually linked with the :attr:`HelmPackageTask.chart_tarball`
    #: output property.
    chart_tarball: Property[Path]

    #: The Helm registry to push to. This can be an HTTP(S) URL to a Helm registry, in which case it must be the full
    #: URL to the remote "directory" where the :attr:`chart_tarball` will be uploaded to. The filename for the chart
    #: in the directory is the basename of :attr:`chart_tarball` unless :attr:`chart_name` is set.
    #:
    #: Alternatively, an `oci://` URL can be specified in which case the `helm push` command is used to push the
    #: chart instead. Note that in this case the :attr:`chart_name` cannot be used and doing so will result in an
    #: error.
    registry: Property[str]

    #: The base name of the chart in the registry. Only when uploading to a HTTP(S) "directory".
    chart_name: Property[str]

    #: The final constructed chart URL that the chart will be published under. Note: This URL is not currently
    #: constructed when pushing to an `oci://` registry and reading the property will cause an error.
    chart_url: Property[str] = Property.output()

    def finalize(self) -> None:
        self.chart_name.setdefault(Supplier.of_callable((lambda: self.chart_tarball.get().name), [self.chart_tarball]))
        return super().finalize()

    def execute(self) -> TaskStatus:
        url = urllib.parse.urlparse(self.registry.get())
        if not url.scheme:
            raise ValueError(f"{self.registry} missing url scheme: {self.registry.get()!r}")
        if url.scheme not in ("oci", "http", "https"):
            raise ValueError(f"{self.registry} invalid url scheme: {self.registry.get()!r}")
        if not url.hostname:
            raise ValueError(f"{self.registry} missing host name: {self.registry.get()!r}")

        settings = helm_settings(self.project)
        credentials: tuple[str, str] | None
        oci_login_host: str | None
        for oci_login_host in (url.hostname, f"{url.hostname}:{url.port}"):
            if oci_login_host in settings.auth:
                credentials = settings.auth[oci_login_host]
                break
        else:
            oci_login_host = None

        if url.scheme == "oci" and credentials:
            self.logger.info("logging into OCI registry %r", url.hostname)
            assert oci_login_host is not None
            command, result = helmapi.helm_registry_login(
                oci_login_host,
                credentials[0],
                credentials[1],
                insecure=oci_login_host in settings.insecure_registries,
            )
            if result != 0:
                return TaskStatus.from_exit_code(command, result)

        if url.scheme == "oci":
            self.chart_url.seterror(f"{self.chart_url} for OCI registries is not currently supported")
            self.logger.info('pushing chart "%s" to OCI registry %r', self.chart_tarball.get(), self.registry.get())
            command, result = helmapi.helm_push(self.chart_tarball.get(), self.registry.get())
            if result != 0:
                return TaskStatus.from_exit_code(command, result)
        elif url.scheme in ("http", "https"):
            self.logger.info(
                'pushing chart "%s" to %s registry %r',
                self.chart_tarball.get(),
                url.scheme.upper(),
                self.registry.get(),
            )
            self.chart_url.set(urllib.parse.urljoin(self.registry.get() + "/", self.chart_name.get()))
            response = httpx.put(self.chart_url.get(), content=self.chart_tarball.get().read_bytes(), auth=credentials)
            response.raise_for_status()
            self.logger.info("chart url = %s", self.chart_url.get())
        else:
            assert False, url

        return TaskStatus.succeeded()



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, List

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class IsortTask(EnvironmentAwareDispatchTask):
    check_only: Property[bool] = Property.config(default=False)
    config_file: Property[Path]
    additional_files: Property[List[Path]] = Property.config(default_factory=list)

    def get_execute_command(self) -> list[str]:
        command = ["isort", "src/"] + self.settings.get_tests_directory_as_args()
        command += [str(p) for p in self.additional_files.get()]
        if self.check_only.get():
            command += ["--check-only"]
        if self.config_file.is_filled():
            command += ["--settings-file", str(self.config_file.get())]
        return command


@dataclasses.dataclass
class IsortTasks:
    check: IsortTask
    format: IsortTask


def isort(project: Project | None = None, **kwargs: Any) -> IsortTasks:
    # TODO (@NiklasRosenstein): We may need to ensure an order to isort and block somehow, sometimes they yield
    #       slightly different results based on the order they run.
    project = project or Project.current()
    check_task = project.do("isortCheck", IsortTask, group="lint", **kwargs, check_only=True)
    format_task = project.do("isortFormat", IsortTask, group="fmt", default=False, **kwargs)
    return IsortTasks(check_task, format_task)



from __future__ import annotations

import dataclasses
import logging
from pathlib import Path

from kraken.core import Project

from .buildsystem import PythonBuildSystem, detect_build_system

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PythonIndex:
    alias: str
    index_url: str
    upload_url: str | None
    credentials: tuple[str, str] | None
    is_package_source: bool
    default: bool


@dataclasses.dataclass
class PythonSettings:
    """Project-global settings for Python tasks."""

    project: Project
    build_system: PythonBuildSystem | None = None
    source_directory: Path = Path("src")
    tests_directory: Path | None = None
    package_indexes: dict[str, PythonIndex] = dataclasses.field(default_factory=dict)
    always_use_managed_env: bool = False

    def get_tests_directory(self) -> Path | None:
        """Returns :attr:`tests_directory` if it is set. If not, it will look for the following directories and
        return the first that exists: `test/`, `tests/`, `src/test/`, `src/tests/`. The determined path will be
        relative to the project directory."""

        if self.tests_directory:
            return self.tests_directory
        for test_dir in map(Path, ["test", "tests", "src/test", "src/tests"]):
            if (self.project.directory / test_dir).is_dir():
                return test_dir
        return None

    def get_tests_directory_as_args(self) -> list[str]:
        """Returns a list with a single item that is the test directory, or an empty list. This is convenient
        when constructing command-line arguments where you want to pass the test directory if it exists."""

        test_dir = self.get_tests_directory()
        return [] if test_dir is None else [str(test_dir)]

    def get_default_package_index(self) -> PythonIndex | None:
        return next((index for index in self.package_indexes.values() if index.default), None)

    def add_package_index(
        self,
        alias: str,
        *,
        index_url: str | None = None,
        upload_url: str | None = None,
        credentials: tuple[str, str] | None = None,
        is_package_source: bool = True,
        default: bool = False,
    ) -> PythonSettings:
        """Adds an index to consume Python packages from or publish packages to.

        :param alias: An alias for the package index.
        :param index_url: The URL of the package index (with the trailing `/simple` bit if applicable).
            If not specified, *alias* must be a known package index (`pypi` or `testpypi`).
        :param upload_url: If the upload url deviates from the registry URL. Otherwise, the upload URL will
            be the same as the *url*,
        :param credentials: Optional credentials to read from the index.
        :param is_package_source: If set to `False`, the index will not be used to source packages from, but
            can be used to publish to.
        """

        if default:
            defidx = self.get_default_package_index()
            if defidx is not None and defidx.alias != alias:
                raise ValueError(f"cannot add another default index (got: {defidx.alias!r}, trying to add: {alias!r})")

        if index_url is None:
            if alias == "pypi":
                index_url = "https://pypi.org/simple"
            elif alias == "testpypi":
                index_url = "https://test.pypi.org/simple"
            else:
                raise ValueError(f"cannot derive index URL for alias {alias!r}")
        if upload_url is None:
            if alias == "pypi":
                upload_url = "https://upload.pypi.org/legacy"
            elif alias == "testpypi":
                upload_url = "https://test.pypi.org/legacy"
            elif index_url.endswith("/simple"):
                upload_url = index_url[: -len("/simple")]
            else:
                raise ValueError(f"cannot derive upload URL for alias {alias!r} and index URL {index_url!r}")

        self.package_indexes[alias] = PythonIndex(alias, index_url, upload_url, credentials, is_package_source, default)
        return self


def python_settings(
    project: Project | None = None,
    build_system: PythonBuildSystem | None = None,
    source_directory: str | Path | None = None,
    tests_directory: str | Path | None = None,
    always_use_managed_env: bool | None = None,
) -> PythonSettings:
    """Read the Python settings for the given or current project and optionally update attributes.

    :param project: The project to get the settings for. If not specified, the current project will be used.
    :environment_handler: If specified, set the :attr:`PythonSettings.environment_handler`. If a string is specified,
        the following values are currently supported: `"poetry"`.
    :param source_directory: The source directory. Defaults to `"src"`.
    :param tests_directory: The tests directory. Automatically determined if left empty.
    """

    project = project or Project.current()
    settings = project.find_metadata(PythonSettings)
    if settings is None:
        settings = PythonSettings(project)
        project.metadata.append(settings)

    if build_system is None and settings.build_system is None:
        # Autodetect the environment handler.
        build_system = detect_build_system(project.directory)
        if build_system:
            logger.info("Detected Python build system %r for %s", type(build_system).__name__, project)

    if build_system is not None:
        if settings.build_system:
            logger.warning(
                "overwriting existing PythonSettings.environment_handler=%r with %r",
                settings.build_system,
                build_system,
            )
        settings.build_system = build_system

    if source_directory is not None:
        settings.source_directory = Path(source_directory)

    if tests_directory is not None:
        settings.tests_directory = Path(tests_directory)

    if always_use_managed_env is not None:
        settings.always_use_managed_env = True

    return settings



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional, Sequence

from kraken.core import Project, Property

from kraken.std.generic.render_file_task import RenderFileTask

from ..gitignore import GitignoreFile, parse_gitignore, sort_gitignore


class GitignoreSyncTask(RenderFileTask):
    """This task ensures that a given set of entries are present in a `.gitignore` file.

    The :attr:`header` property can be set to place the paths below a particular comment in the `.gitignore` file. If
    there is no comment with the given text, it and the paths will be appended to the end of the file. When no header
    is specified, only missing paths will be added to beginning of the `.gitignore` file.

    If :attr:`sort` is enabled, the `.gitignore` file will be sorted (keeping paths grouped under their comments).

    It's common to group this task under the default `fmt` group, as it is similar to formatting a `.gitignore` file.
    """

    file: Property[Path] = Property.default(".gitignore")
    sort_paths: Property[bool] = Property.config(default=False)
    sort_groups: Property[bool] = Property.config(default=False)

    _paths: Dict[Optional[str], List[str]]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self._paths = {}

    def add_paths(self, header: str | None, paths: Sequence[str]) -> None:
        self._paths.setdefault(header, []).extend(paths)

    # RenderFileTask

    def get_file_contents(self, file: Path) -> str | bytes:
        if file.exists():
            gitignore = parse_gitignore(file)
        else:
            gitignore = GitignoreFile([])

        for comment_header, paths in self._paths.items():
            has_paths = set(gitignore.paths())

            if comment_header is not None:

                # Remove all existing paths, we'll make sure they're located under the header.
                for path in paths:
                    if path in has_paths:
                        gitignore.remove_path(path)
                        has_paths.discard(path)

                # Find the location of the header.
                insert_index = gitignore.find_comment(comment_header)
                if insert_index is None:
                    gitignore.add_blank()
                    gitignore.add_comment(comment_header)
                    insert_index = len(gitignore.entries)
                else:
                    insert_index += 1

            else:
                insert_index = 0

            for path in paths:
                if path not in has_paths:
                    gitignore.add_path(path, insert_index)
                    insert_index += 1

        gitignore = sort_gitignore(gitignore, self.sort_paths.get(), self.sort_groups.get())

        return gitignore.render()



from __future__ import annotations

import dataclasses
import re
import subprocess as sp
from pathlib import Path


def git_describe(path: Path | None, tags: bool = True, dirty: bool = True) -> str:
    """Describe a repository with tags.

    :param path: The directory in which to describe.
    :param tags: Whether to include tags (adds the `--tags` flag).
    :param dirty: Whether to include if the directory tree is dirty (adds the `--dirty` flag).
    :raise ValueError: If `git describe` failed.
    :return: The Git head description.
    """

    command = ["git", "describe"]
    if tags:
        command.append("--tags")
    if dirty:
        command.append("--dirty")
    try:
        return sp.check_output(command, cwd=path).decode().strip()
    except sp.CalledProcessError:
        raise ValueError("could not describe Git repository")


@dataclasses.dataclass
class GitVersion:
    """Represents a "git version" that has a major, minor and patch version and optionall a commit distance."""

    @dataclasses.dataclass
    class CommitDistance:
        value: int
        sha: str

    major: int
    minor: int
    patch: int
    distance: CommitDistance | None
    dirty: bool

    @staticmethod
    def parse(value: str) -> GitVersion:
        GIT_VERSION_REGEX = r"^(\d+)\.(\d+)\.(\d+)(?:-(\d+)-g(\w+))?(-dirty)?$"
        match = re.match(GIT_VERSION_REGEX, value)
        if not match:
            raise ValueError(f"not a valid GitVersion: {value!r}")
        if match.group(4):
            distance = GitVersion.CommitDistance(value=int(match.group(4)), sha=match.group(5))
        else:
            distance = None
        return GitVersion(
            major=int(match.group(1)),
            minor=int(match.group(2)),
            patch=int(match.group(3)),
            distance=distance,
            dirty=match.group(6) is not None,
        )



from __future__ import annotations

from pathlib import Path
from typing import Any, List

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class Flake8Task(EnvironmentAwareDispatchTask):
    config_file: Property[Path]
    additional_args: Property[List[str]] = Property.config(default_factory=list)

    def get_execute_command(self) -> list[str]:
        command = ["flake8", "src/"] + self.settings.get_tests_directory_as_args()
        if self.config_file.is_filled():
            command += ["--config", str(self.config_file.get())]
        command += self.additional_args.get()
        return command


def flake8(project: Project | None = None, **kwargs: Any) -> Flake8Task:
    project = project or Project.current()
    return project.do("flake8", Flake8Task, group="lint", **kwargs)



import logging
from pathlib import Path
from typing import Dict, List

from kraken.core import Project, Property
from kraken.core.utils import not_none

from ..config import CargoRegistry
from .cargo_build_task import CargoBuildTask

logger = logging.getLogger(__name__)


class CargoPublishTask(CargoBuildTask):
    """Publish a Cargo crate."""

    #: Path to the Cargo configuration file (defaults to `.cargo/config.toml`).
    cargo_config_file: Property[Path] = Property.default(".cargo/config.toml")

    #: The registry to publish the package to.
    registry: Property[CargoRegistry]

    def get_cargo_command(self, env: Dict[str, str]) -> List[str]:
        super().get_cargo_command(env)
        registry = self.registry.get()
        if registry.publish_token is None:
            raise ValueError(f'registry {registry.alias!r} missing a "publish_token"')
        return (
            ["cargo", "publish"]
            + self.additional_args.get()
            + ["--registry", registry.alias, "--token", registry.publish_token]
        )

    def make_safe(self, args: List[str], env: Dict[str, str]) -> None:
        args[args.index(not_none(self.registry.get().publish_token))] = "[MASKED]"
        super().make_safe(args, env)

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self._base_command = ["cargo", "publish"]



from __future__ import annotations

import dataclasses

from kraken.core import Project


@dataclasses.dataclass
class CargoRegistry:
    """Represents a Cargo registry."""

    #: The registrt alias. This is used as an identifier when publishing the registry and when referencing a crate
    #: from the registry in the `Cargo.toml` dependencies.
    alias: str

    #: The URL of the Cargo registry index. This usually points to a Git repository, as that is how Cargo registries
    #: are stored. The index URL must be present in `.cargo/config.toml` for Cargo to consume crates from it.
    index: str

    #: Authentication credentials for reading from the registry. This is only needed if the registry is private and the
    #: index URL is an HTTP(S) URL. The credentials will be passed using HTTP Basic authentication.
    read_credentials: tuple[str, str] | None = None

    #: The publish token for this registry.
    publish_token: str | None = None


@dataclasses.dataclass
class CargoProject:
    """Container for all Cargo related settings that can be automatically managed from a Kraken build."""

    #: The registries for the Cargo project. We store the registrie's by their alias.
    registries: dict[str, CargoRegistry] = dataclasses.field(default_factory=dict)

    #: Environment variables for cargo build steps.
    build_env: dict[str, str] = dataclasses.field(default_factory=dict)

    def add_registry(
        self,
        alias: str,
        index: str,
        read_credentials: tuple[str, str] | None = None,
        publish_token: str | None = None,
    ) -> None:
        """Add a registry to the project.

        :param alias: The alias of the registry. This alias is used in` Cargo.toml` to describe which registry to look
            up a create in. It is also used to designate the registry to publish to in `cargo publish`.
        :param index: The registry index URL.
        :param read_credentials: A `(username, password)` tuple for reading from the repository (optional).
        :param publish_token: A token to publish to the repository (optional).
        """

        self.registries[alias] = CargoRegistry(alias, index, read_credentials, publish_token)

    @staticmethod
    def get_or_create(project: Project | None) -> CargoProject:
        project = project or Project.current()
        return project.find_metadata(CargoProject, CargoProject)



from __future__ import annotations

from ..git import GitVersion


def git_version_to_python_version(value: str | GitVersion, include_sha: bool) -> str:
    """Converts a Git version to a Python version.

    :param value: The Git version to convert.
    :param sha: Include the SHA of the commit distance if it exists.
    """

    version = GitVersion.parse(value) if isinstance(value, str) else value
    final_version = f"{version.major}.{version.minor}.{version.patch}"
    if version.distance:
        final_version += f".dev{version.distance.value}"
        if include_sha:
            final_version += f"+g{version.distance.sha}"
            if version.dirty:
                final_version += "-dirty"
    if version.dirty and "-dirty" not in final_version:
        final_version += "+dirty"
    return final_version



""" Abstraction of Python build systems such as Poetry and Slap. """


from __future__ import annotations

import abc
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..settings import PythonSettings


class PythonBuildSystem(abc.ABC):
    """Abstraction of a Python build system."""

    @abc.abstractmethod
    def supports_managed_environments(self) -> bool:
        """Return `True` if the build system supports managed environments."""

    @abc.abstractmethod
    def get_managed_environment(self) -> ManagedEnvironment:
        """Return a handle for the managed environment.

        :raise NotImplementedError: If :meth:`supports_managed_environment` returns `False`.
        """

    @abc.abstractmethod
    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        """Build one or more distributions of the project managed by this build system.

        :param output_directory: The directory where the distributions should be placed.
        :param as_version: A version number for the built distributions.
        """


class ManagedEnvironment(abc.ABC):
    """Abstraction of a managed Python environment."""

    @abc.abstractmethod
    def exists(self) -> bool:
        """Return `True` if the managed environment exists."""

    @abc.abstractmethod
    def get_path(self) -> Path:
        """Return the path to the managed environment.

        :raises RuntimeError: May be raised if the environment does not exist (it may not be possible to
            determine the path of the environment before it exists depending on the build system).
        """

    @abc.abstractmethod
    def install(self, settings: PythonSettings) -> None:
        """Install the managed environment. This should be a no-op if the environment already exists."""


def detect_build_system(project_directory: Path) -> PythonBuildSystem | None:
    """Detect the Python build system used in *project_directory*."""

    pyproject_toml = project_directory / "pyproject.toml"
    if not pyproject_toml.is_file():
        return None

    pyproject_content = pyproject_toml.read_text()

    if "[tool.slap]" in pyproject_content:
        from .slap import SlapPythonBuildSystem

        return SlapPythonBuildSystem(project_directory)

    if "poetry-core" in pyproject_content:
        from .poetry import PoetryPythonBuildSystem

        return PoetryPythonBuildSystem(project_directory)

    return None



from __future__ import annotations

import base64
import contextlib
import json
import shlex
import tempfile
from pathlib import Path

from kraken.core.project import Project
from kraken.core.property import Property
from kraken.core.utils import flatten

from . import DockerBuildTask
from .dockerapi import docker_load, docker_run


class KanikoBuildTask(DockerBuildTask):
    """An implementation for building Docker images with Kaniko."""

    kaniko_image: Property[str]
    kaniko_context: Property[str]
    kaniko_cache_copy_layers: Property[bool]
    kaniko_snapshot_mode: Property[str]
    kaniko_secrets_mount_dir: Property[str]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.kaniko_image.set("gcr.io/kaniko-project/executor:debug")
        self.kaniko_context.set("/workspace")
        self.kaniko_cache_copy_layers.set(True)
        self.kaniko_snapshot_mode.set("redo")
        self.kaniko_secrets_mount_dir.set("/run/secrets")

    def finalize(self) -> None:
        if self.cache.get() and not self.push.get() and not self.cache_repo.get():
            self.logger.warning(
                "Disabling cache in Kaniko build %s because it must be combined with push or cache_repo",
                self,
            )
            self.cache.set(False)
        cache_repo = self.cache_repo.get_or(None)
        if cache_repo and ":" in cache_repo:
            raise ValueError(f"Kaniko --cache-repo argument cannot contain `:` (got: {cache_repo!r})")
        return super().finalize()

    def _render_auth_file(self) -> str:
        """Renders the JSON configuration that will be written to `/kaniko/.docker/config.json`."""
        return json.dumps(
            {
                "auths": {
                    index: {"auth": base64.b64encode(f"{username}:{password}".encode("ascii")).decode("ascii")}
                    for index, (username, password) in self.auth.get().items()
                }
            },
            indent=2,
        )

    def _render_main_script(self, executor_command: list[str]) -> str:
        """Renders the shell script that will be executed in the Kaniko container."""

        docker_config = self._render_auth_file()

        script = []
        script += [
            "mkdir -p /kaniko/.docker",
            "cat << EOF > /kaniko/.docker/config.json",
            docker_config,
            "EOF",
        ]

        if self.secrets:
            script += [f"mkdir -p {shlex.quote(self.kaniko_secrets_mount_dir.get())}"]
            for secret, value in self.secrets.get().items():
                script += [
                    f"echo {shlex.quote(value)} > {shlex.quote(self.kaniko_secrets_mount_dir.get() + '/' + secret)}"
                ]

        script += [" ".join(map(shlex.quote, executor_command))]
        return "\n".join(script)

    def _get_kaniko_executor_command(self, dockerfile: str | None, tar_path: str | None) -> list[str]:
        if tar_path and not self.tags:
            raise ValueError("Need at least one destination (tag) when exporting to an image tarball")
        executor_command = ["/kaniko/executor"]
        executor_command += flatten(("--build-arg", f"{key}={value}") for key, value in self.build_args.get().items())
        cache_repo = self.cache_repo.get()
        if cache_repo:
            executor_command += ["--cache-repo", cache_repo]
        if self.cache.get_or(False):
            executor_command += ["--cache=true"]
        executor_command += flatten(("--destination", destination) for destination in self.tags.get())
        if dockerfile:
            executor_command += ["--dockerfile", dockerfile]
        if not self.push.get():
            executor_command += ["--no-push"]
        executor_command += ["--snapshotMode", self.kaniko_snapshot_mode.get()]
        if self.squash.get():
            executor_command += ["--single-snapshot"]
        if self.kaniko_cache_copy_layers.get():
            executor_command += ["--cache-copy-layers"]
        target = self.target.get()
        if target:
            executor_command += ["--target", target]
        if tar_path:
            executor_command += ["--tarPath", tar_path]
        executor_command += ["--context", self.kaniko_context.get()]
        return executor_command

    def _build(
        self,
        exit_stack: contextlib.ExitStack,
    ) -> None:
        volumes = [f"{self.build_context.get().absolute()}:{self.kaniko_context.get()}"]

        # If the Dockerfile is not relative to the build context, we need to mount it explicitly.
        in_container_dockerfile: str | None = None
        if self.dockerfile.is_filled():
            try:
                in_container_dockerfile = str(
                    self.dockerfile.get().absolute().relative_to(self.build_context.get().absolute())
                )
            except ValueError:
                in_container_dockerfile = "/kaniko/Dockerfile"
                volumes += [f"{self.dockerfile.get().absolute()}:{in_container_dockerfile}"]

        # If the image needs to be loaded into the Docker daemon after building, we need to always
        # export it to a file.
        image_output_file = self.image_output_file.get_or(None)
        if self.load.get() and not image_output_file:
            tempdir = exit_stack.enter_context(tempfile.TemporaryDirectory())
            image_output_file = Path(tempdir) / "image.tgz"

        # Construct the tar path for inside the container.
        tar_path: str | None = None
        if image_output_file:
            volumes += [f"{image_output_file.parent.absolute()}:/kaniko/out"]
            tar_path = f"/kaniko/out/{image_output_file.name}"

        executor_command = self._get_kaniko_executor_command(in_container_dockerfile, tar_path)

        script = self._render_main_script(executor_command)

        result = docker_run(
            image=self.kaniko_image.get(),
            args=["sh", "-c", script],
            entrypoint="",
            remove=True,
            volumes=volumes,
            workdir=self.kaniko_context.get(),
            platform=self.platform.get_or(None),
        )

        if result != 0:
            raise Exception(f"Kaniko build failed with exit code {result}")

        if self.load.get():
            assert image_output_file is not None, "image_output_file is expected to be set when config.load == True"
            result = docker_load(image_output_file)
            if result != 0:
                raise Exception(f"Docker load failed with exit code {result}")

    def execute(self) -> None:
        with contextlib.ExitStack() as exit_stack:
            self._build(exit_stack)



from .settings import PythonSettings, python_settings
from .tasks.black_task import BlackTask, black
from .tasks.build_task import BuildTask, build
from .tasks.flake8_task import Flake8Task, flake8
from .tasks.install_task import InstallTask, install
from .tasks.isort_task import IsortTask, isort
from .tasks.mypy_task import MypyTask, mypy
from .tasks.publish_task import PublishTask, publish
from .tasks.pytest_task import PytestTask, pytest

# Backwards compatibilty
from .version import git_version_to_python_version, git_version_to_python_version as git_version_to_python

__all__ = [
    "PythonSettings",
    "python_settings",
    "BlackTask",
    "black",
    "BuildTask",
    "build",
    "InstallTask",
    "install",
    "IsortTask",
    "isort",
    "Flake8Task",
    "flake8",
    "MypyTask",
    "mypy",
    "PublishTask",
    "publish",
    "PytestTask",
    "pytest",
    "git_version_to_python_version",
    "git_version_to_python",
]



from __future__ import annotations

import logging
from pathlib import Path
from typing import List, Optional

from kraken.core import Project, Property, Supplier, Task, TaskStatus

from ..buildsystem import PythonBuildSystem
from ..settings import python_settings

logger = logging.getLogger(__name__)


class BuildTask(Task):
    build_system: Property[Optional[PythonBuildSystem]]
    output_directory: Property[Path]
    as_version: Property[Optional[str]] = Property.config(default=None)
    output_files: Property[List[Path]] = Property.output()

    def execute(self) -> TaskStatus:
        build_system = self.build_system.get()
        if not build_system:
            return TaskStatus.failed("no build system configured")
        output_directory = self.output_directory.get_or(self.project.build_directory / "python-dist")
        output_directory.mkdir(exist_ok=True, parents=True)
        self.output_files.set(build_system.build(output_directory, self.as_version.get()))
        return TaskStatus.succeeded()


def build(
    *,
    name: str = "pythonBuild",
    group: str | None = "build",
    as_version: str | None = None,
    project: Project | None = None,
) -> BuildTask:
    """Creates a build task for the given project.

    The build task relies on the build system configured in the Python project settings."""

    project = project or Project.current()
    task = project.do(
        name,
        BuildTask,
        default=False,
        group=group,
        build_system=Supplier.of_callable(lambda: python_settings(project).build_system),
        as_version=as_version,
    )
    return task



from __future__ import annotations

import contextlib
from pathlib import Path

import tomli
import tomli_w
from kraken.core import BackgroundTask, Property, TaskStatus
from kraken.core.utils import atomic_file_swap


class CargoBumpVersionTask(BackgroundTask):
    """This task bumps the version number in `Cargo.toml`. The change can be reverted afterwards if the
    :attr:`revert` option is enabled."""

    version: Property[str]
    revert: Property[bool] = Property.default(False)
    cargo_toml_file: Property[Path] = Property.default("Cargo.toml")

    def _get_updated_cargo_toml(self) -> str:
        content = tomli.loads(self.cargo_toml_file.get().read_text())
        content["package"]["version"] = self.version.get()
        return tomli_w.dumps(content)

    # BackgroundTask

    def start_background_task(self, exit_stack: contextlib.ExitStack) -> TaskStatus | None:
        content = self._get_updated_cargo_toml()
        revert = self.revert.get()
        fp = exit_stack.enter_context(atomic_file_swap(self.cargo_toml_file.get(), "w", always_revert=revert))
        fp.write(content)
        fp.close()
        version = self.version.get()
        return (
            TaskStatus.started(f"temporary bump to {version}")
            if revert
            else TaskStatus.succeeded(f"permanent bump to {version}")
        )

    # Task

    def finalize(self) -> None:
        self.cargo_toml_file.set(self.cargo_toml_file.value.map(lambda path: self.project.directory / path))



from __future__ import annotations

from kraken.std.git import GitVersion


def git_version_to_cargo_version(version: str | GitVersion, include_sha: bool) -> str:
    """Constructs a Cargo version from a Git version."""

    version = GitVersion.parse(version) if isinstance(version, str) else version
    cargo_version = f"{version.major}.{version.minor}.{version.patch}"
    if version.distance:
        cargo_version = f"{cargo_version}-dev{version.distance.value}"
        if include_sha:
            cargo_version = f"{cargo_version}+{version.distance.sha}"
    return cargo_version



from __future__ import annotations

import enum
import io
from itertools import islice
from os import PathLike
from pathlib import Path
from typing import Iterable, NamedTuple, TextIO


class GitignoreEntryType(enum.Enum):
    COMMENT = enum.auto()
    BLANK = enum.auto()
    PATH = enum.auto()


class GitignoreEntry(NamedTuple):
    type: GitignoreEntryType
    value: str

    def __str__(self) -> str:
        if self.is_comment():
            return f"# {self.value}"
        return self.value

    def is_comment(self) -> bool:
        return self.type == GitignoreEntryType.COMMENT

    def is_blank(self) -> bool:
        return self.type == GitignoreEntryType.BLANK

    def is_path(self) -> bool:
        return self.type == GitignoreEntryType.PATH


class GitignoreFile(NamedTuple):
    entries: list[GitignoreEntry]

    def find_comment(self, comment: str) -> int | None:
        return next(
            (i for i, e in enumerate(self.entries) if e.is_comment() and e.value.lstrip("#").strip() == comment), None
        )

    def paths(self, start: int | None = None, stop: int | None = None) -> Iterable[str]:
        return (entry.value for entry in islice(self.entries, start, stop) if entry.is_path())

    def add_comment(self, comment: str, index: int | None = None) -> None:
        entry = GitignoreEntry(GitignoreEntryType.COMMENT, comment)
        self.entries.insert(len(self.entries) if index is None else index, entry)

    def add_blank(self, index: int | None = None) -> None:
        entry = GitignoreEntry(GitignoreEntryType.BLANK, "")
        self.entries.insert(len(self.entries) if index is None else index, entry)

    def add_path(self, path: str, index: int | None = None) -> None:
        entry = GitignoreEntry(GitignoreEntryType.PATH, path)
        self.entries.insert(len(self.entries) if index is None else index, entry)

    def remove_path(self, path: str) -> None:
        index = next((i for i, e in enumerate(self.entries) if e.is_path() and e.value == path), None)
        if index is None:
            raise ValueError(f'"{path}" not in GitignoreFile')
        del self.entries[index]

    def render(self) -> str:
        return "\n".join(map(str, self.entries)) + "\n"


def parse_gitignore(file: TextIO | Path | str) -> GitignoreFile:
    if isinstance(file, str):
        return parse_gitignore(io.StringIO(file))
    elif isinstance(file, PathLike):
        with file.open() as fp:
            return parse_gitignore(fp)

    result = GitignoreFile([])
    for line in file:
        line = line.rstrip("\n")
        if line.startswith("#"):
            line = line[1:].lstrip()
            type = GitignoreEntryType.COMMENT
        elif not line.strip():
            line = ""
            type = GitignoreEntryType.BLANK
        else:
            type = GitignoreEntryType.PATH
        result.entries.append(GitignoreEntry(type, line))

    return result


def sort_gitignore(gitignore: GitignoreFile, sort_paths: bool = True, sort_groups: bool = False) -> GitignoreFile:
    """Sorts the entries in the specified gitignore file, keeping paths under a common comment block grouped.
    Will also get rid of any extra blanks.

    :param gitignore: The input to sort.
    :param sort_paths: Whether to sort paths (default: True).
    :param sort_groups: Whether to sort groups among themselves, not just paths within groups (default: False).
    :return: A new, sorted gitignore file.
    """

    class Group(NamedTuple):
        comments: list[str]
        paths: list[str]

    # List of (comments, paths).
    groups: list[Group] = [Group([], [])]

    for entry in gitignore.entries:
        if entry.is_path():
            groups[-1].paths.append(entry.value)
        elif entry.is_comment():
            # If we already have paths in the current group, we open a new group.
            if groups[-1].paths:
                groups.append(Group([entry.value], []))
            # Otherwise we append the comment to the group.
            else:
                groups[-1].comments.append(entry.value)

    if sort_groups:
        groups.sort(key=lambda g: "\n".join(g.comments).lower())

    new = GitignoreFile([])
    for group in groups:
        if sort_paths:
            group.paths.sort(key=str.lower)
        for comment in group.comments:
            new.add_comment(comment)
        for path in group.paths:
            new.add_path(path)
        new.add_blank()

    if new.entries and new.entries[-1].is_blank():
        new.entries.pop()

    return new



import subprocess as sp

from kraken.core import Property, Task, TaskStatus


class CargoFmtTask(Task):
    check: Property[bool] = Property.default(False)

    def execute(self) -> TaskStatus:
        command = ["cargo", "fmt"]
        if self.check.get():
            command += ["--check"]
        return TaskStatus.from_exit_code(command, sp.call(command, cwd=self.project.directory))



""" Provides tasks for Rust projects that build using Cargo. """

from __future__ import annotations

from typing import Sequence

from kraken.core import Project, Supplier
from typing_extensions import Literal

from .config import CargoProject, CargoRegistry
from .tasks.cargo_auth_proxy_task import CargoAuthProxyTask
from .tasks.cargo_build_task import CargoBuildTask
from .tasks.cargo_bump_version_task import CargoBumpVersionTask
from .tasks.cargo_clippy_task import CargoClippyTask
from .tasks.cargo_fmt_task import CargoFmtTask
from .tasks.cargo_publish_task import CargoPublishTask
from .tasks.cargo_sync_config_task import CargoSyncConfigTask

__all__ = [
    "cargo_auth_proxy",
    "cargo_build",
    "cargo_clippy",
    "cargo_fmt",
    "cargo_publish",
    "cargo_registry",
    "cargo_sync_config",
    "CargoAuthProxyTask",
    "CargoBuildTask",
    "CargoBumpVersionTask",
    "CargoClippyTask",
    "CargoProject",
    "CargoPublishTask",
    "CargoRegistry",
    "CargoSyncConfigTask",
]

CARGO_SYNC_CONFIG_TASK_NAME = "cargoSyncConfig"
CARGO_AUTH_PROXY_TASK_NAME = "cargoAuthProxy"


def cargo_registry(
    alias: str,
    index: str,
    read_credentials: tuple[str, str] | None = None,
    publish_token: str | None = None,
    project: Project | None = None,
) -> None:
    """Adds a Cargo registry to the project. The registry must be synced to disk into the `.cargo/config.toml`
    configuration file. You need to make sure to add a sync task using :func:`cargo_sync_config` if you manage
    your Cargo registries with this function. Can be called multiple times.

    :param alias: The registry alias.
    :param index: The registry index URL (usually an HTTPS URL that ends in `.git`).
    :param read_credentials: Username/password to read from the registry (only for private registries).
    :param publish_token: The token to use with `cargo publish`.

    !!! note Artifactory

        It appears that for Artifactory, the *publish_token* must be of the form `Bearer <TOKEN>` where the token
        is a token generated manually via the JFrog UI. It cannot be an API key.
    """

    cargo = CargoProject.get_or_create(project)
    cargo.add_registry(alias, index, read_credentials, publish_token)


def cargo_auth_proxy(*, project: Project | None = None) -> CargoAuthProxyTask:
    """Creates a background task that the :func:`cargo_build` and :func:`cargo_publish` tasks will depend on to
    inject the read credentials for private registries into HTTPS requests made by Cargo. This is only needed when
    private registries are used."""

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)
    task = project.do(
        CARGO_AUTH_PROXY_TASK_NAME,
        CargoAuthProxyTask,
        False,
        registries=Supplier.of_callable(lambda: list(cargo.registries.values())),
    )
    task.add_relationship(f":{CARGO_SYNC_CONFIG_TASK_NAME}?")
    return task


def cargo_sync_config(
    *,
    project: Project | None = None,
    replace: bool = False,
) -> CargoSyncConfigTask:
    """Creates a task that the :func:`cargo_build` and :func:`cargo_publish` tasks will depend on to synchronize
    the `.cargo/config.toml` configuration file, ensuring that the Cargo registries configured with the
    :func:`cargo_registry` function are present and up to date."""

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)
    return project.do(
        CARGO_SYNC_CONFIG_TASK_NAME,
        CargoSyncConfigTask,
        True,
        group="fmt",
        registries=Supplier.of_callable(lambda: list(cargo.registries.values())),
        replace=replace,
    )


def cargo_clippy(
    *,
    fix: bool = False,
    name: str | None = None,
    group: str | None = "_auto_",
    project: Project | None = None,
) -> None:
    project = project or Project.current()
    name = "cargoClippyFix" if fix else "cargoClippy"
    group = ("fmt" if fix else "lint") if group == "_auto_" else group
    project.do(name, CargoClippyTask, not fix, group=group, fix=fix)


def cargo_fmt(*, project: Project | None = None) -> None:
    project = project or Project.current()
    project.do("cargoFmt", CargoFmtTask, False, group="fmt")
    project.do("cargoFmtCheck", CargoFmtTask, True, group="lint", check=True)


def cargo_build(
    mode: Literal["debug", "release"],
    incremental: bool | None = None,
    env: dict[str, str] | None = None,
    *,
    group: str | None = "build",
    name: str | None = None,
    project: Project | None = None,
) -> CargoBuildTask:
    """Creates a task that runs `cargo build`.

    :param mode: Whether to create a task that runs the debug or release build.
    :param incremental: Whether to build incrementally or not (with the `--incremental=` option). If not
        specified, the option is not specified and the default behaviour is used.
    :param env: Override variables for the build environment variables. Values in this dictionary override
        variables in :attr:`CargoProject.build_env`.
    :param name: The name of the task. If not specified, defaults to `:cargoBuild{mode.capitalied()}`."""

    assert mode in ("debug", "release"), repr(mode)
    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)
    task = project.do(
        f"cargoBuild{mode.capitalize()}" if name is None else name,
        CargoBuildTask,
        default=False,
        group=group,
        incremental=incremental,
        args=["--release"] if mode == "release" else [],
        env=Supplier.of_callable(lambda: {**cargo.build_env, **(env or {})}),
    )
    task.add_relationship(f":{CARGO_AUTH_PROXY_TASK_NAME}?")
    task.add_relationship(f":{CARGO_SYNC_CONFIG_TASK_NAME}?")
    return task


def cargo_publish(
    registry: str,
    incremental: bool | None = None,
    env: dict[str, str] | None = None,
    *,
    version: str | None = None,
    additional_args: Sequence[str] = (),
    name: str = "cargoPublish",
    project: Project | None = None,
) -> CargoPublishTask:
    """Creates a task that publishes the create to the specified *registry*.

    :param registry: The alias of the registry to publish to.
    :param incremental: Incremental builds on or off.
    :param env: Environment variables (overrides :attr:`CargoProject.build_env`).
    :param version: The version number to publish. If specified, a cargo bump task will be added. If a version
        number to bump to is specified, the `--allow-dirty` option is automatically passed to Cargo.
    """

    project = project or Project.current()
    cargo = CargoProject.get_or_create(project)

    additional_args = list(additional_args)
    if version is not None and "--allow-dirty" not in additional_args:
        additional_args.append("--allow-dirty")

    task = project.do(
        name,
        CargoPublishTask,
        False,
        group="publish",
        registry=Supplier.of_callable(lambda: cargo.registries[registry]),
        additional_args=additional_args,
        incremental=incremental,
        env=Supplier.of_callable(lambda: {**cargo.build_env, **(env or {})}),
    )
    task.add_relationship(f":{CARGO_AUTH_PROXY_TASK_NAME}?")
    task.add_relationship(f":{CARGO_SYNC_CONFIG_TASK_NAME}?")

    if version is not None:
        bump_task = project.do(
            f"{name}/bump",
            CargoBumpVersionTask,
            False,
            version=version,
            revert=True,
        )
        task.add_relationship(bump_task)

    return task



from __future__ import annotations

from pathlib import Path
from typing import Iterable, List

import tomli
import tomli_w
from kraken.core import Property, Task, TaskRelationship

from kraken.std.generic.render_file_task import RenderFileTask

from ..config import CargoRegistry


class CargoSyncConfigTask(RenderFileTask):
    """This task updates the `.cargo/config.toml` file to inject configuration values."""

    # Override the default value of the :attr:`RenderFileTask.file`.
    file: Property[Path] = Property.default(".cargo/config.toml")

    #: If enabled, the configuration file will be replaced rather than updated.
    replace: Property[bool] = Property.config(default=False)

    #: The registries to insert into the configuration.
    registries: Property[List[CargoRegistry]] = Property.config(default_factory=list)

    #: Tasks that are dependant on this task.
    for_tasks: Property[List[Task]] = Property.default_factory(list)

    # RenderFileTask

    def get_file_contents(self, file: Path) -> str | bytes:
        content = tomli.loads(file.read_text()) if not self.replace.get() and file.exists() else {}
        for registry in self.registries.get():
            content.setdefault("registries", {})[registry.alias] = {"index": registry.index}
        lines = []
        if self.replace.get():
            lines.append("# This file is managed by Kraken. Manual edits to this file will be overwritten.")
        else:
            lines.append(
                "# This file is partially managed by Kraken. Comments and manually added "
                "repositories are not preserved."
            )
        lines.append(tomli_w.dumps(content))
        return "\n".join(lines)

    # Task

    def get_relationships(self) -> Iterable[TaskRelationship]:
        yield from super().get_relationships()
        for task in self.for_tasks.get():
            yield TaskRelationship(task, True, True)



from __future__ import annotations

from pathlib import Path
from typing import Any, List

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class MypyTask(EnvironmentAwareDispatchTask):
    config_file: Property[Path]
    additional_args: Property[List[str]] = Property.config(default_factory=list)
    check_tests: Property[bool] = Property.config(default=True)
    use_daemon: Property[bool] = Property.config(default=True)
    python_version: Property[str]

    def get_execute_command(self) -> list[str]:
        # TODO (@NiklasRosenstein): Should we somewhere add a task that ensures `.dmypy.json` is in `.gitignore`?
        #       Having it in the project directory makes it easier to just stop the daemon if it malfunctions (which
        #       happens regularly but is hard to detect automatically).
        status_file = self.project.directory / ".dmypy.json"
        command = ["dmypy", "--status-file", str(status_file), "run", "--"] if self.use_daemon.get() else ["mypy"]
        if self.config_file.is_filled():
            command += ["--config-file", str(self.config_file.get())]
        else:
            command += ["--show-error-codes", "--namespace-packages"]  # Sane defaults. 🙏
        if self.python_version.is_filled():
            command += ["--python-version", self.python_version.get()]
        source_dir = self.settings.source_directory
        command += [str(source_dir)]
        if self.check_tests.get():
            # We only want to add the tests directory if it is not already in the source directory. Otherwise
            # Mypy will find the test files twice and error.
            tests_dir = self.settings.get_tests_directory()
            if tests_dir:
                try:
                    tests_dir.relative_to(source_dir)
                except ValueError:
                    command += [str(tests_dir)]
        command += self.additional_args.get()
        return command


def mypy(project: Project | None = None, **kwargs: Any) -> MypyTask:
    project = project or Project.current()
    return project.do("mypy", MypyTask, group="lint", **kwargs)



from __future__ import annotations

import os
import shlex
from pathlib import Path
from typing import Any, List

from kraken.core import Project, Property, TaskStatus
from kraken.core.utils import flatten

from .base_task import EnvironmentAwareDispatchTask

# TODO (@NiklasRosenstein): Pytest coverage support


class PytestTask(EnvironmentAwareDispatchTask):
    tests_dir: Property[Path]
    ignore_dirs: Property[List[Path]] = Property.config(default_factory=list)
    allow_no_tests: Property[bool] = Property.config(default=False)

    def is_skippable(self) -> bool:
        return self.allow_no_tests.get() and self.tests_dir.is_empty() and not self.settings.get_tests_directory()

    def get_execute_command(self) -> list[str] | TaskStatus:
        tests_dir = self.tests_dir.get_or(None)
        tests_dir = tests_dir or self.settings.get_tests_directory()
        if not tests_dir:
            print("error: no test directory configured and none could be detected")
            return TaskStatus.failed("no test directory configured and none could be detected")
        command = ["pytest", "-vv", str(self.project.directory / tests_dir)]
        command += flatten(["--ignore", str(self.project.directory / path)] for path in self.ignore_dirs.get())
        command += ["--log-cli-level", "INFO"]
        command += shlex.split(os.getenv("PYTEST_FLAGS", ""))
        return command

    def handle_exit_code(self, code: int) -> TaskStatus:
        if code == 5 and self.allow_no_tests.get():
            # Pytest returns exit code 5 if no tests were run.
            return TaskStatus.succeeded()
        return TaskStatus.from_exit_code(None, code)


def pytest(*, name: str = "pytest", group: str = "test", project: Project | None = None, **kwargs: Any) -> PytestTask:
    project = project or Project.current()
    return project.do(name, PytestTask, group=group, **kwargs)



from __future__ import annotations

from pathlib import Path

from kraken.core import Project, Property, Task, TaskStatus

DEFAULT_ENCODING = "utf-8"


class RenderFileTask(Task):
    file: Property[Path]
    encoding: Property[str]
    content: Property[str]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.encoding.setdefault(DEFAULT_ENCODING)
        self._content_cache: bytes | None = None

    def get_file_contents(self, file: Path) -> str | bytes:
        return self.content.get()

    def finalize(self) -> None:
        self.file.set(self.file.value.map(lambda p: self.project.directory / p))
        super().finalize()

    def prepare(self) -> TaskStatus | None:
        file = self.file.get()

        # Materialize the file contents.
        content = self.get_file_contents(file)
        if isinstance(content, str):
            self._content_cache = content.encode(self.encoding.get())
        else:
            self._content_cache = content

        # Check if we would be updating the file.
        if file.is_file() and file.read_bytes() == self._content_cache:
            return TaskStatus.up_to_date()

        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        assert self._content_cache is not None
        file = self.file.get()
        file.parent.mkdir(exist_ok=True)
        file.write_bytes(self._content_cache)
        return TaskStatus.succeeded(f"write {len(self._content_cache)} bytes to {file}")



""" Implements a MITM proxy server using the :mod:`proxy` (`proxy.py` on PyPI) module to inject the auth credentials
into Cargo and Git HTTP(S) requests. """

from __future__ import annotations

import base64
import contextlib
import json
import logging
import os
import subprocess as sp
import sys
from pathlib import Path
from typing import Iterator, Optional

from kraken.core.utils import not_none
from proxy.http.parser import HttpParser
from proxy.http.proxy.plugin import HttpProxyBasePlugin

logger = logging.getLogger(__name__)


class AuthInjector(HttpProxyBasePlugin):
    """This proxy.py plugin injects credentials according to the `INJECT_AUTH` environment variable."""

    # TODO (@NiklasRosenstein): Find a better way than environment variables to pass configuration to this plugin.

    _auth: dict[str, tuple[str, str]] | None = None

    @property
    def auth(self) -> dict[str, tuple[str, str]]:
        if self._auth is None:
            self._auth = json.loads(os.environ["INJECT_AUTH"])
        return self._auth

    def handle_client_request(self, request: HttpParser) -> Optional[HttpParser]:

        # NOTE (@NiklasRosenstein): This method is only called for requests in an HTTPS tunnel if TLS
        #       interception is enabled, which requires a self-signed CA-certificate.

        host = not_none(request.headers)[b"host"][1].partition(b":")[0].decode()
        # path = request.path.decode() if request.path else None

        # if path and ('http://' not in path or 'https://' not in path):
        #     # For the proxied HTTPS requests, it appears that we only get the path instead of the full URL in
        #     # the request path, so we make sure we
        #     new_url = f'https://{host}{path}'
        #     request.set_url(new_url.encode())

        if host in self.auth and not request.has_header(b"Authorization"):
            logger.info("injecting Authorization for %s request to %s", not_none(request.method).decode(), host)
            creds = self.auth[host]
            auth = base64.b64encode(f"{creds[0]}:{creds[1]}".encode())
            request.add_header(b"Authorization", b"Basic " + auth)

        return request


@contextlib.contextmanager
def mitm_auth_proxy(
    auth: dict[str, tuple[str, str]],
    port: int = 8899,
    timeout: float | None = None,
) -> Iterator[tuple[str, Path]]:
    """Runs a MITM HTTPS proxy that injects credentials according to *auth* into requests."""

    if timeout is None and "PROXY_PY_TIMEOUT" in os.environ:
        timeout = float(os.environ["PROXY_PY_TIMEOUT"])

    certs_dir = Path(__file__).parent / "data" / "certs"
    key_file = certs_dir / "key.pem"
    cert_file = certs_dir / "cert.pem"

    command = [
        sys.executable,
        "-m",
        "proxy",
        "--plugins",
        __name__ + "." + AuthInjector.__name__,
        "--ca-key-file",
        str(key_file),
        "--ca-cert-file",
        str(cert_file),
        "--ca-signing-key",
        str(key_file),
        "--port",
        str(port),
    ]

    if timeout is not None:
        command += ["--timeout", str(timeout)]

    env = os.environ.copy()
    env["INJECT_AUTH"] = json.dumps(auth)

    logger.info("starting proxy server: %s", command)
    proc = sp.Popen(command, env=env)

    try:
        yield f"http://localhost:{port}", cert_file
    finally:
        logger.info("stopping proxy server")
        proc.terminate()
        proc.wait()
        if proc.returncode is None:
            proc.kill()
            proc.wait()



""" Tools for Git versioned projects. """

from __future__ import annotations

from typing import Optional, Sequence, cast

from kraken.core import Project

from .tasks.gitignore_sync_task import GitignoreSyncTask
from .version import GitVersion, git_describe

__all__ = [
    "git_describe",
    "GitVersion",
    "GitignoreSyncTask",
    "gitignore",
]

GITIGNORE_TASK_NAME = "gitignore"


def gitignore(header: str | None, paths: Sequence[str] | str, *, project: Project | None = None) -> GitignoreSyncTask:
    project = project or Project.current()
    task = cast(Optional[GitignoreSyncTask], project.tasks().get(GITIGNORE_TASK_NAME))
    if task is None:
        task = project.do(GITIGNORE_TASK_NAME, GitignoreSyncTask, group="fmt")
    task.add_paths(header, [paths] if isinstance(paths, str) else paths)
    return task



from __future__ import annotations

import contextlib
import logging
import time
from pathlib import Path
from typing import Iterable, Iterator, List
from urllib.parse import urlparse

import tomli
import tomli_w
from kraken.core import BackgroundTask, Property, Task, TaskRelationship
from kraken.core.utils import atomic_file_swap, not_none

from kraken.std.cargo.config import CargoRegistry
from kraken.std.git.config import dump_gitconfig, load_gitconfig

logger = logging.getLogger(__name__)


class CargoAuthProxyTask(BackgroundTask):
    """This task starts a local proxy server that injects HTTP basic authentication credentials in HTTP(S) requests
    to a Cargo repository to work around Cargo's current inability to interface with private repositories."""

    #: The Cargo config file to update.
    cargo_config_file: Property[Path] = Property.default(".cargo/config.toml")

    #: A list of the Cargo registries for which to inject credentials for based on matching paths.
    registries: Property[List[CargoRegistry]]

    #: The proxy port.
    proxy_port: Property[int] = Property.default(8899)

    #: The URL of the proxy. This property is only valid and accessible for tasks that immediately directly depend
    #: on this task and will be invalidated at the task teardown.
    proxy_url: Property[str] = Property.output()

    #: The path to the certificate file that needs to be trusted in order to talk to the proxy over HTTPS.
    proxy_cert_file: Property[Path] = Property.output()

    #: Tasks that are dependant on this task.
    for_tasks: Property[List[Task]] = Property.default_factory(list)

    @contextlib.contextmanager
    def _inject_config(self) -> Iterator[None]:
        """Injects the proxy URL and cert file into the Cargo and Git configuration."""

        # TODO (@NiklasRosenstein): Can we get away without temporarily modifying the GLOBAL Git config?

        cargo_config_toml = self.project.directory / self.cargo_config_file.get()
        cargo_config = tomli.loads(cargo_config_toml.read_text()) if cargo_config_toml.is_file() else {}

        git_config_file = Path("~/.gitconfig").expanduser()
        git_config = load_gitconfig(git_config_file) if git_config_file.is_file() else {}

        with contextlib.ExitStack() as exit_stack:
            # Temporarily update the Cargo configuration file to inject the HTTP(S) proxy and CA info.
            cargo_http = cargo_config.setdefault("http", {})
            cargo_http["proxy"] = self.proxy_url.get()
            cargo_http["cainfo"] = str(self.proxy_cert_file.get().absolute())
            logger.info("updating %s", cargo_config_toml)
            fp = exit_stack.enter_context(
                atomic_file_swap(cargo_config_toml, "w", always_revert=True, create_dirs=True)
            )
            fp.write(tomli_w.dumps(cargo_config))
            fp.close()

            # Temporarily update the Git configuration file to inject the HTTP(S) proxy and CA info.
            git_http = git_config.setdefault("http", {})
            git_http["proxy"] = self.proxy_url.get()
            git_http["sslCAInfo"] = str(self.proxy_cert_file.get().absolute())
            logger.info("updating %s", git_config_file)
            fp = exit_stack.enter_context(atomic_file_swap(git_config_file, "w", always_revert=True, create_dirs=True))
            fp.write(dump_gitconfig(git_config))
            fp.close()

            yield

    # Task

    def get_relationships(self) -> Iterable[TaskRelationship]:
        yield from super().get_relationships()
        for task in self.for_tasks.get():
            yield TaskRelationship(task, True, True)

    def start_background_task(self, exit_stack: contextlib.ExitStack) -> None:
        from ..mitm import mitm_auth_proxy

        auth: dict[str, tuple[str, str]] = {}
        for registry in self.registries.get():
            if not registry.read_credentials:
                continue
            host = not_none(urlparse(registry.index).hostname)
            auth[host] = registry.read_credentials

        proxy_url, cert_file = exit_stack.enter_context(mitm_auth_proxy(auth=auth, port=self.proxy_port.get()))
        self.proxy_url.set(proxy_url)
        exit_stack.callback(lambda: self.proxy_url.clear())
        self.proxy_cert_file.set(cert_file)
        exit_stack.callback(lambda: self.proxy_cert_file.clear())
        time.sleep(1)  # Give the proxy some time to start up.

        exit_stack.enter_context(self._inject_config())



from __future__ import annotations

import contextlib
import shutil
import subprocess as sp
import tempfile
from pathlib import Path


def helm_package(
    chart_path: Path,
    output_file: Path | None = None,
    output_directory: Path | None = None,
    app_version: str | None = None,
    version: str | None = None,
) -> tuple[int, Path | None]:
    """Package a Helm chart."""

    if output_file is not None and output_directory is not None:
        raise ValueError("output_file and output_directory cannot both be set")

    with contextlib.ExitStack() as exit_stack:
        command = ["helm", "package", str(chart_path)]

        tempdir: Path | None = None
        if output_directory is None or output_file is not None:
            # We build into a temporary directory first.
            tempdir = Path(exit_stack.enter_context(tempfile.TemporaryDirectory()))
            command += ["--destination", str(tempdir)]
        else:
            command += ["--destination", str(output_directory)]
        if app_version:
            command += ["--appVersion", app_version]
        if version:
            command += ["--version", version]

        result = sp.call(command)
        if result != 0:
            return result, None

        if output_file:
            assert tempdir is not None
            output_file.parent.mkdir(exist_ok=True, parents=True)
            if output_file.exists():
                output_file.unlink()
            chart_file = next(Path(tempdir).iterdir())
            shutil.move(str(chart_file), output_file)
        else:
            assert output_directory is not None
            chart_file = next(output_directory.iterdir())

        return 0, chart_file

    assert False


def helm_registry_login(registry: str, username: str, password: str, insecure: bool = False) -> tuple[list[str], int]:
    """Log into a Helm registry."""

    command = ["helm", "registry", "login", registry, "-u", username, "--password-stdin"]
    if insecure:
        command += ["--insecure"]
    return command, sp.run(command, input=f"{password}\n".encode()).returncode


def helm_push(chart_tarball: Path, remote: str) -> tuple[list[str], int]:
    """Push a Helm chart to a remote."""

    command = ["helm", "push", str(chart_tarball), remote]
    return command, sp.call(command)



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, List, Union

from kraken.core import Project, Property

from .base_task import EnvironmentAwareDispatchTask


class BlackTask(EnvironmentAwareDispatchTask):
    """A task to run the `black` formatter to either check for necessary changes or apply changes."""

    check_only: Property[bool] = Property.config(default=False)
    config_file: Property[Path]
    source_directories: Property[List[Union[str, Path]]] = Property.config(default_factory=lambda: ["src"])
    additional_args: Property[List[str]] = Property.config(default_factory=list)
    additional_files: Property[List[Path]] = Property.config(default_factory=list)

    def get_execute_command(self) -> list[str]:
        command = ["black"] + list(map(str, self.source_directories.get()))
        command += self.settings.get_tests_directory_as_args()
        command += [str(p) for p in self.additional_files.get()]
        if self.check_only.get():
            command += ["--check"]
        if self.config_file.is_filled():
            command += ["--config", str(self.config_file.get())]
        command += self.additional_args.get()
        return command


@dataclasses.dataclass
class BlackTasks:
    check: BlackTask
    format: BlackTask


def black(project: Project | None = None, **kwargs: Any) -> BlackTasks:
    """Creates two black tasks, one to check and another to format. The check task will be grouped under `"lint"`
    whereas the format task will be grouped under `"fmt"`."""

    project = project or Project.current()
    check_task = project.do("blackCheck", BlackTask, group="lint", **kwargs, check_only=True)
    format_task = project.do("blackFormat", BlackTask, group="fmt", default=False, **kwargs)
    return BlackTasks(check_task, format_task)



from __future__ import annotations

import contextlib
import dataclasses
import os
import re
import shutil
import subprocess as sp
from pathlib import Path
from typing import Any

from kraken.core import BackgroundTask, Project, Property, TaskStatus


@dataclasses.dataclass
class AzureBlobStorageCache:
    connection_string: str
    container_name: str
    key_prefix: str | None = None

    def to_env(self) -> dict[str, str]:
        environ = {
            "SCCACHE_AZURE_CONNECTION_STRING": self.connection_string,
            "SCCACHE_AZURE_BLOB_CONTAINER": self.container_name,
        }
        if self.key_prefix is not None:
            environ["SCCACHE_AZURE_KEY_PREFIX"] = self.key_prefix
        return environ


@dataclasses.dataclass
class LocalCache:
    cache_dir: Path | None = None

    def to_env(self) -> dict[str, str]:
        environ = {}
        if self.cache_dir:
            environ["SCCACHE_DIR"] = str(self.cache_dir.absolute())
        return environ


@dataclasses.dataclass
class SccacheManager:
    cache_config: AzureBlobStorageCache | LocalCache | None
    log_level: str | None = None
    log_file: Path | None = None
    bin: Path | None = None

    def __post_init__(self) -> None:
        self._proc: sp.Popen[Any] | None = None

    def is_running(self) -> bool:
        return self._proc is not None and self._proc.returncode is None

    def get_cache_location(self) -> str:
        stats = self.stats()
        match = re.search(r"^Cache\s+location\s+(.*)$", stats, re.I | re.MULTILINE)
        assert match is not None, f"Could not determine cache location from sccache stats output: {stats!r}"
        return match.group(1)

    def start(self) -> None:
        """Start the Sccache server."""

        if self.is_running():
            assert self._proc is not None
            raise RuntimeError(f"Sccache is already running (pid: {self._proc.pid})")

        command = [str(self.bin) if self.bin else "sccache"]
        env = {}
        if self.cache_config:
            env.update(self.cache_config.to_env())
        if self.log_level is not None:
            env["SCCACHE_LOG"] = self.log_level
        if self.log_file is not None:
            env["SCCACHE_ERROR_LOG"] = str(self.log_file.absolute())
        env["SCCACHE_START_SERVER"] = "1"
        self._proc = sp.Popen(command, env={**os.environ, **env})

    def stats(self) -> str:
        command = [str(self.bin) if self.bin else "sccache", "-s"]
        return sp.check_output(command).decode()

    def stop(self) -> None:
        """Stop the Sccache server."""

        if not self.is_running():
            return
        assert self._proc is not None

        env = {"SCCACHE_NO_DAEMON": "1"}
        command = [str(self.bin) if self.bin else "sccache", "--stop-server"]
        sp.check_call(command, env={**os.environ, **env})

        self._proc.wait(10)
        if self.is_running():
            self._proc.terminate()


def find_sccache() -> Path | None:
    sccache = shutil.which("sccache")
    if sccache is not None:
        return Path(sccache)
    return None


class SccacheTask(BackgroundTask):
    """This task ensures that an Sccache server is running for all its dependant tasks."""

    manager: Property[SccacheManager]

    def start_background_task(self, exit_stack: contextlib.ExitStack) -> TaskStatus:
        manager = self.manager.get()
        if not manager.is_running():
            manager.start()
        exit_stack.callback(lambda: print(manager.stats()))
        exit_stack.callback(lambda: manager.stop())
        return TaskStatus.started(manager.get_cache_location())


def sccache(manager: SccacheManager, *, name: str = "sccache", project: Project | None = None) -> SccacheTask:
    """Creates a background task that starts the sccache server."""

    project = project or Project.current()
    return project.do(name, SccacheTask, False, manager=manager)



from __future__ import annotations

from pathlib import Path
from typing import Iterable, List, Optional, Tuple

from kraken.core import Project, Property, Task, TaskRelationship
from twine.commands.upload import upload as twine_upload
from twine.settings import Settings as TwineSettings

from ..settings import python_settings


class PublishTask(Task):
    """Publishes Python distributions to one or more indexes using :mod:`twine`."""

    index_upload_url: Property[str]
    index_credentials: Property[Optional[Tuple[str, str]]] = Property.config(default=None)
    distributions: Property[List[Path]]
    skip_existing: Property[bool] = Property.config(default=False)
    dependencies: List[Task]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.dependencies = []

    def get_relationships(self) -> Iterable[TaskRelationship]:
        yield from (TaskRelationship(task, True, False) for task in self.dependencies)
        yield from super().get_relationships()

    def execute(self) -> None:
        credentials = self.index_credentials.get()
        settings = TwineSettings(
            repository_url=self.index_upload_url.get().rstrip("/") + "/",
            username=credentials[0] if credentials else None,
            password=credentials[1] if credentials else None,
            skip_existing=self.skip_existing.get(),
            non_interactive=True,
        )
        twine_upload(settings, list(map(str, self.distributions.get())))


def publish(
    *,
    package_index: str,
    distributions: list[Path] | Property[List[Path]],
    skip_existing: bool = False,
    name: str = "publishPython",
    group: str | None = "publish",
    default: bool = False,
    project: Project | None = None,
    after: list[Task] | None = None,
) -> PublishTask:
    """Create a publish task for the specified registry."""

    project = project or Project.current()
    settings = python_settings(project)
    if package_index not in settings.package_indexes:
        raise ValueError(f"package index {package_index!r} is not defined")

    index = settings.package_indexes[package_index]
    task = project.do(
        name,
        PublishTask,
        default=default,
        group=group,
        index_upload_url=index.upload_url,
        index_credentials=index.credentials,
        distributions=distributions,
        skip_existing=skip_existing,
    )
    task.dependencies += after or []
    return task



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from kraken.core import Project, Property, Task
from kraken.core.utils import import_class

__version__ = "0.1.0"
__all__ = ["build", "DockerBuildTask", "manifest_tool"]

DEFAULT_BUILD_BACKEND = "native"
BUILD_BACKENDS = {
    "buildx": f"{__name__}.buildx.BuildxBuildTask",
    "kaniko": f"{__name__}.kaniko.KanikoBuildTask",
    "native": f"{__name__}.native.NativeBuildTask",
}


class DockerBuildTask(Task):
    """Base class for tasks that build Docker images. Subclasses implement converting the task properties into
    the invokation for a Docker build backend."""

    build_context: Property[Path]
    dockerfile: Property[Path]
    auth: Property[Dict[str, Tuple[str, str]]]
    platform: Property[str]
    build_args: Property[Dict[str, str]]
    secrets: Property[Dict[str, str]]
    cache_repo: Property[Optional[str]]
    cache: Property[bool]
    tags: Property[List[str]]
    push: Property[bool]
    squash: Property[bool]
    target: Property[Optional[str]]
    image_output_file: Property[Optional[Path]]
    load: Property[bool]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.build_context.set(project.directory)
        self.auth.setdefault({})
        self.build_args.setdefault({})
        self.secrets.setdefault({})
        self.cache_repo.setdefault(None)
        self.cache.set(True)
        self.tags.setdefault([])
        self.push.set(False)
        self.squash.set(False)
        self.target.setdefault(None)
        self.image_output_file.setdefault(None)
        self.load.setdefault(False)


def build_docker_image(
    *,
    name: str = "buildDocker",
    backend: str = DEFAULT_BUILD_BACKEND,
    project: Project | None = None,
    **kwds: Any,
) -> DockerBuildTask:
    """Create a new task in the current project that builds a Docker image and eventually pushes it."""

    task_class = import_class(BUILD_BACKENDS[backend], DockerBuildTask)  # type: ignore[misc]
    return (project or Project.current()).do(name, task_class, **kwds)


from .manifest_tool import manifest_tool  # noqa: E402



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['std',
 'std.cargo',
 'std.cargo.tasks',
 'std.docker',
 'std.generic',
 'std.git',
 'std.git.tasks',
 'std.helm',
 'std.python',
 'std.python.buildsystem',
 'std.python.tasks']

package_data = \
{'': ['*'], 'std.cargo': ['data/certs/*']}

install_requires = \
['httpx>=0.23.0,<0.24.0',
 'kraken-core>=0.5.4,<0.6.0',
 'nr.python.environment>=0.1.4,<0.2.0',
 'proxy.py>=2.4.3,<3.0.0',
 'tomli-w>=1.0.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0',
 'twine>=4.0.1,<5.0.0']

setup_kwargs = {
    'name': 'kraken-std',
    'version': '0.1.33',
    'description': 'The Kraken standard library.',
    'long_description': '# kraken-std\n\n[![Python application](https://github.com/kraken-build/kraken-std/actions/workflows/python-package.yml/badge.svg)](https://github.com/kraken-build/kraken-std/actions/workflows/python-package.yml)\n[![PyPI version](https://badge.fury.io/py/kraken-std.svg)](https://badge.fury.io/py/kraken-std)\n\nThe Kraken standard library.\n\n---\n\n## Development\n\n### Integration testing\n\nIntegration tests are located in `src/tests/integration`. The following tools need to be available to run the\nintegration tests:\n\n* Cargo (to test Cargo building and publishing) *The Cargo integration tests run against Artifactory and Cloudsmith\nand requires credentials to temporarily create a new Cargo repository (available in CI).*\n* Docker (used to setup services that we run integration tests against)\n* Helm (to test Helm packaging and publishing)\n* Poetry (to test Python publishing and installing)\n* [Slap](https://github.com/python-slap/slap-cli) (to test Python publishing and installing)\n\n__Test a single integration test__\n\n    ```\n    PYTEST_FLAGS="--log-cli-level DEBUG -s -k <test_filter>" kraken run pytestIntegration -v\n    ```\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



""" Implements Poetry as a build system for kraken-std. """

from __future__ import annotations

import contextlib
import logging
import os
import shutil
import subprocess as sp
from pathlib import Path

import tomli
import tomli_w
from kraken.core.utils import NotSet, atomic_file_swap, is_relative_to
from nr.python.environment.virtualenv import get_current_venv

from kraken.std.python.settings import PythonSettings

from . import ManagedEnvironment, PythonBuildSystem

logger = logging.getLogger(__name__)


class PoetryPythonBuildSystem(PythonBuildSystem):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory

    def supports_managed_environments(self) -> bool:
        return True

    def get_managed_environment(self) -> ManagedEnvironment:
        return PoetryManagedEnvironment(self.project_directory)

    def build(self, output_directory: Path, as_version: str | None = None) -> list[Path]:
        if as_version is not None:
            # TODO (@NiklasRosenstein): We should find a way to revert the changes to the worktree
            #       that this command does.
            command = ["poetry", "version", as_version]
            logger.info("%s", command)
            sp.check_call(command, cwd=self.project_directory)

        # Poetry does not allow configuring the output folder, so it's always going to be "dist/".
        # We remove the contents of that folder to make sure we know what was produced.
        dist_dir = self.project_directory / "dist"
        if dist_dir.exists():
            shutil.rmtree(dist_dir)

        command = ["poetry", "build"]
        logger.info("%s", command)
        sp.check_call(command, cwd=self.project_directory)
        src_files = list(dist_dir.iterdir())
        dst_files = [output_directory / path.name for path in src_files]
        for src, dst in zip(src_files, dst_files):
            shutil.move(str(src), dst)

        # Unless the output directory is a subdirectory of the dist_dir, we remove the dist dir again.
        if not is_relative_to(output_directory, dist_dir):
            shutil.rmtree(dist_dir)

        return dst_files


class PoetryManagedEnvironment(ManagedEnvironment):
    def __init__(self, project_directory: Path) -> None:
        self.project_directory = project_directory
        self._env_path: Path | None | NotSet = NotSet.Value

    def exists(self) -> bool:
        try:
            self.get_path()
            return True
        except RuntimeError:
            return False

    def get_path(self) -> Path:
        if self._env_path is NotSet.Value:
            command = ["poetry", "env", "info", "-p"]
            environ = os.environ.copy()
            # Poetry would otherwise assume the active virtual env.
            venv = get_current_venv(environ)
            if venv:
                venv.deactivate(environ)
            try:
                self._env_path = Path(
                    sp.check_output(
                        command,
                        cwd=self.project_directory,
                        env=environ,
                    )
                    .decode()
                    .strip()
                )
            except sp.CalledProcessError as exc:
                if exc.returncode != 1:
                    raise
                self._env_path = None
        if self._env_path is None:
            raise RuntimeError("managed environment does not exist")
        return self._env_path

    def install(self, settings: PythonSettings) -> None:

        # Ensure that `poetry.toml` is up to date with the credentials.
        poetry_toml = self.project_directory / "poetry.toml"
        poetry_conf = tomli.loads(poetry_toml.read_text()) if poetry_toml.exists() else {}

        # Ensure that the source is configured in `pyproject.toml`.
        # TODO (@NiklasRosenstein): Maybe we should permanently sync the source configuration into the
        #       pyproject.toml instead of just temporarily?
        pyproject_toml = self.project_directory / "pyproject.toml"
        pyproject_conf = tomli.loads(pyproject_toml.read_text())

        for index in settings.package_indexes.values():
            if index.is_package_source and index.credentials:
                poetry_conf.setdefault("http-basic", {})[index.alias] = {
                    "username": index.credentials[0],
                    "password": index.credentials[1],
                }
                pyproject_conf.setdefault("tool", {}).setdefault("poetry", {}).setdefault("source", []).append(
                    {
                        "name": index.alias,
                        "url": index.index_url,
                        "secondary": True,
                    }
                )

        with contextlib.ExitStack() as exit_stack:
            fp = exit_stack.enter_context(atomic_file_swap(poetry_toml, "wb", always_revert=True))
            tomli_w.dump(poetry_conf, fp)
            fp.close()

            fp = exit_stack.enter_context(atomic_file_swap(pyproject_toml, "wb", always_revert=True))
            tomli_w.dump(pyproject_conf, fp)
            fp.close()

            command = ["poetry", "install", "--no-interaction"]
            logger.info("%s", command)
            sp.check_call(command, cwd=self.project_directory)



from __future__ import annotations

import abc
import logging
import os
import subprocess as sp
from typing import Iterable, MutableMapping

from kraken.core import Project, Task, TaskRelationship, TaskStatus
from nr.python.environment.virtualenv import VirtualEnvInfo, get_current_venv

from kraken.std.python.buildsystem import ManagedEnvironment

from ..settings import python_settings

logger = logging.getLogger(__name__)


class EnvironmentAwareDispatchTask(Task):
    """Base class for tasks that run a subcommand. The command ensures that the command is aware of the
    environment configured in the project settings."""

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.settings = python_settings(project)

    def get_relationships(self) -> Iterable[TaskRelationship]:
        # If a pythonInstall task exists, we may need it.
        install_task = self.project.tasks().get("pythonInstall")
        if install_task:
            yield TaskRelationship(install_task, True, False)
        yield from super().get_relationships()

    @abc.abstractmethod
    def get_execute_command(self) -> list[str] | TaskStatus:
        pass

    def handle_exit_code(self, code: int) -> TaskStatus:
        return TaskStatus.from_exit_code(None, code)

    def activate_managed_environment(self, venv: ManagedEnvironment, environ: MutableMapping[str, str]) -> None:
        active_venv = get_current_venv(environ)
        if active_venv is None or self.settings.always_use_managed_env:
            if not venv.exists():
                logger.warning("Managed environment (%s) does not exist", venv)
                return
            managed_env = VirtualEnvInfo(venv.get_path())
            logger.info("Activating managed environment (%s)", managed_env.path)
            managed_env.activate(environ)
        elif active_venv:
            logger.info("An active virtual environment was found, not activating managed environment")

    def execute(self) -> TaskStatus:
        command = self.get_execute_command()
        if isinstance(command, TaskStatus):
            return command
        env = os.environ.copy()
        if self.settings.build_system and self.settings.build_system.supports_managed_environments():
            self.activate_managed_environment(self.settings.build_system.get_managed_environment(), env)
        logger.info("%s", command)
        result = sp.call(command, cwd=self.project.directory, env=env)
        return self.handle_exit_code(result)



from __future__ import annotations

import configparser
import io
from pathlib import Path
from typing import TextIO


def load_gitconfig(file: TextIO | Path | str) -> dict[str, dict[str, str]]:
    """Parses a Git configuration file."""

    if isinstance(file, str):
        return load_gitconfig(io.StringIO(file))
    elif isinstance(file, Path):
        with file.open() as fp:
            return load_gitconfig(fp)

    parser = configparser.RawConfigParser()
    parser.read_file(file)
    result = dict(parser._sections)  # type: ignore[attr-defined]
    for k in result:
        result[k] = dict(parser._defaults, **result[k])  # type: ignore[attr-defined]
        result[k].pop("__name__", None)
    return result


def dump_gitconfig(data: dict[str, dict[str, str]]) -> str:
    """Formats a Git configuration file."""

    parser = configparser.RawConfigParser()
    for section, values in data.items():
        parser.add_section(section)
        for key, value in values.items():
            parser.set(section, key, value)
    fp = io.StringIO()
    parser.write(fp)
    return fp.getvalue()



from __future__ import annotations

import subprocess as sp
from typing import List, Sequence

from kraken.core import Project, Property, Task, TaskStatus

RELEASE_URL = (
    "https://github.com/estesp/manifest-tool/releases/download/v{VERSION}/binaries-manifest-tool-{VERSION}.tar.gz"
)

# TODO (@NiklasRosenstein): Use existing manifest tool if it exists.
# TODO (@NiklasRosenstein): Ensure manifest-tool has credentials to push to the target


class ManifestToolPushTask(Task):
    """A task that uses `manifest-tool` to combine multiple container images from different platforms into a single
    multi-platform manifest.

    For more information on `manifest-tool`, check out the GitHub repository:

    https://github.com/estesp/manifest-tool/
    """

    #: The Docker platforms to create the manifest for.
    platforms: Property[List[str]]

    #: A Docker image tag that should contain the variables `OS`, `ARCH` and `VARIANT`.
    template: Property[str]

    #: The image ID to push the Docker image to.
    target: Property[str]

    def execute(self) -> TaskStatus:
        command = [
            "manifest-tool",
            "push",
            "from-args",
            "--platforms",
            ",".join(self.platforms.get()),
            "--template",
            self.template.get(),
            "--target",
            self.target.get(),
        ]
        self.logger.info("%s", command)
        result = sp.call(command)
        return TaskStatus.from_exit_code(command, result)


def manifest_tool(
    *,
    name: str,
    template: str,
    platforms: Sequence[str],
    target: str,
    inputs: Sequence[Task],
    group: str | None = None,
    project: Project | None = None,
) -> ManifestToolPushTask:
    project = Project.current()
    task = project.do(name, ManifestToolPushTask, group=group, template=template, target=target, platforms=platforms)
    task.add_relationship(inputs)
    return task



import os
import subprocess as sp
from typing import Dict, List, Optional

from kraken.core import Project, Property, Task, TaskStatus


class CargoBuildTask(Task):
    """This task runs `cargo build` using the specified parameters. It will respect the authentication
    credentials configured in :attr:`CargoProjectSettings.auth`."""

    #: Additional arguments to pass to the Cargo command-line.
    additional_args: Property[List[str]] = Property.default_factory(list)

    #: Whether to build incrementally or not.
    incremental: Property[Optional[bool]] = Property.default(None)

    #: Environment variables for the Cargo command.
    env: Property[Dict[str, str]] = Property.default_factory(dict)

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)

    def get_cargo_command(self, env: Dict[str, str]) -> List[str]:
        incremental = self.incremental.get()
        if incremental is not None:
            env["CARGO_INCREMENTAL"] = "1" if incremental else "0"

        return ["cargo", "build"] + self.additional_args.get()

    def make_safe(self, args: List[str], env: Dict[str, str]) -> None:
        pass

    def execute(self) -> TaskStatus:
        env = self.env.get()
        command = self.get_cargo_command(env)

        safe_command = command[:]
        safe_env = env.copy()
        self.make_safe(safe_command, safe_env)
        self.logger.info("%s [env: %s]", safe_command, safe_env)

        result = sp.call(command, cwd=self.project.directory, env={**os.environ, **env})
        return TaskStatus.from_exit_code(command, result)
